/*
 * Copyright (C) 2017-2021 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <iostream>
#include <assert.h>
#include "proba-flooding-ring-routing-agent.h"
#include "../../scheduler.h"
#include<cmath>

//==============================================================================
//
//          ProbaFloodingDynamicRingRoutingAgent  (class)
//
//==============================================================================

mt19937_64 *ProbaFloodingDynamicRingRoutingAgent::forwardingRNG =  new mt19937_64(2);
set<int> ProbaFloodingDynamicRingRoutingAgent::reachability = set<int>();

ProbaFloodingDynamicRingRoutingAgent::ProbaFloodingDynamicRingRoutingAgent(Node *_hostNode) : RoutingAgent(_hostNode) {
  needToSendControl = true;
  rangeBig = ScenarioParameters::getRangeBig();
  //rangeSmall = ScenarioParameters::getRangeSmall();
  initialisationStarted = false;
  forwardingProbability = (ScenarioParameters::getProbability()) * hostNode->getNeighboursCount() / hostNode->getRingNeighboursCount(ScenarioParameters::getRangeSmall(),ScenarioParameters::getRangeBig()); //probability for the ring and not the neighbors //TODO NOT THIS RANGESMALL
}

ProbaFloodingDynamicRingRoutingAgent::~ProbaFloodingDynamicRingRoutingAgent() {
  aliveAgents--;

  if ( aliveAgents == 0 ){
    cout << " Proba flooding with dynamic ring:: number of nodes reached : " << reachability.size() << endl;
    cout << " Proba flooding with dynamic ring:: forwarded packets       : " << forwardedDataPackets << endl;
  }
}

void ProbaFloodingDynamicRingRoutingAgent::initializeAgent() {
  readRangeAttributes ();
}

//receivePacketFromApplication
void ProbaFloodingDynamicRingRoutingAgent::receivePacketFromApplication(PacketPtr _packet) { //on source  
  if (_packet->type == PacketType::D1_DENSITY_PROBE || _packet->type == PacketType::D1_DENSITY_INIT){ //on all nodes approx
    hostNode->enqueueOutgoingPacket(_packet); 
  }
  else{
    float optimal_NbForwardersPerHop = 15.0; 
    float ratio_RingNeighborsOverForwarders = 4.0;
    float optimal_NbRingNeighborsPerHop = optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders;
    if ( hostNode->getEstimatedNeighbours() > optimal_NbRingNeighborsPerHop )

	rangeSmall = sqrt ((- optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders * pow(ScenarioParameters::getRangeBig(),2) / hostNode->getEstimatedNeighbours()) + pow(ScenarioParameters::getRangeBig(),2));

    else rangeSmall = 0;
     
    cout << hostNode->getId() << " deden " << hostNode->getEstimatedNeighbours() << " rangeSmall " << rangeSmall << endl;

    if (needToSendControl) { //before the very first data transmission
      PacketPtr controlBig(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
      controlBig->setCommunicationRange(rangeBig);
      controlBig->setCtrl1SequenceNumber(hostNode->getNextCtrl1SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlBig);
      PacketPtr controlSmall(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
      controlSmall->setCommunicationRange(rangeSmall);
      controlSmall->setCtrl2SequenceNumber(hostNode->getNextCtrl2SequenceNumber());
      hostNode->enqueueOutgoingPacket(controlSmall);	
      needToSendControl = false;													     
    }
    _packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
    hostNode->enqueueOutgoingPacket(_packet);   
    //cout << " estimated " << hostNode->getEstimatedNeighbours() << " real " << hostNode->getNeighboursCount() << endl;

    datapacket_t datapacket;
    datapacket.first =  _packet->srcId;
    datapacket.second = _packet->srcSequenceNumber;    
    auto rangeIt = countAckMap.find(datapacket); //or datapacket, node computes rangeSmall for each new source, or for each new packet?
    if (rangeIt == countAckMap.end()) { 												       countAckMap.insert(pair<datapacket_t,distance_t>(datapacket,rangeSmall));
      auto timerEvent = new ProbaFloodingDynamicRingAckEvent(Scheduler::now(), this, _packet); 
      Scheduler::getScheduler().schedule(timerEvent);
    }
  }
}

//receivePacketFromNetwork
void ProbaFloodingDynamicRingRoutingAgent::receivePacketFromNetwork(PacketPtr _packet) {
  //if ( _packet->srcId == hostNode->getId()){
  //return;} //TODO DO WE NEED IT?

  //DATA
  uniform_real_distribution<float> distrib(0, 1);
  float forwardingProbabilityNumber = distrib(*forwardingRNG);

  if (_packet->type == PacketType::DATA){
    auto srcIt = alreadySeenPackets.find(_packet->srcId);
    if (srcIt == alreadySeenPackets.end() || srcIt->second < _packet->srcSequenceNumber){
      if (srcIt != alreadySeenPackets.end()) {  // already seen
	if (srcIt->second < _packet->srcSequenceNumber) {
	  // the source sequence number of this packet is higher than the highest already seen from this source
	  srcIt->second = _packet->srcSequenceNumber;
	}
      } 
      else {  // seen for the first time
	alreadySeenPackets.insert(pair<int,int>(_packet->srcId, _packet->srcSequenceNumber));
      }

      if(forwardingProbabilityNumber < forwardingProbability && amIOnRing(_packet)){
	forwardDataPackets(_packet);
      }
    }//if not already seen
  }//if data
  
  //CONTROL_BIG
  else if (_packet->type == PacketType::CONTROL_1) {
    auto trIt = amIonRingMap.find(_packet->transmitterId);
    if (trIt != amIonRingMap.end()) {// transmiter already seen
      // the source sequence number of this packet is higher than the highest already seen from this source
      trIt->second.controlSeqNoBig = _packet->ctrl1SequenceNumber;
    } 
    else {
      // transmitter seen for the first time
      contSeqNo_t control;
      control.controlSeqNoBig = _packet->ctrl1SequenceNumber;
      control.controlSeqNoSmall = -1;
      amIonRingMap.insert(pair<int,contSeqNo_t>(_packet->transmitterId, control));
    }	
  }

  //CONTROL_SMALL
  else if (_packet->type == PacketType::CONTROL_2) {
    auto trIt = amIonRingMap.find(_packet->transmitterId);
    if (trIt != amIonRingMap.end()) {// already seen
      // the source sequence number of this packet is higher than the highest already seen from this source
      trIt->second.controlSeqNoSmall = _packet->ctrl2SequenceNumber;
    } 
    else {
      // seen for the first time
      contSeqNo_t control;
      control.controlSeqNoSmall = _packet->ctrl2SequenceNumber;
      control.controlSeqNoBig = -1;
      amIonRingMap.insert(pair<int,contSeqNo_t>(_packet->transmitterId, control));	
    }	
  }

  else if ( _packet->type == PacketType::D1_DENSITY_PROBE) {  //PROBE is local
    hostNode->dispatchPacketToApplication(_packet);
    return;
  }
  else if (  _packet->type == PacketType::D1_DENSITY_INIT && !initialisationStarted ) { //INIT is global
    initialisationStarted = true;
    hostNode->dispatchPacketToApplication(_packet);
    PacketPtr newPacket(_packet->clone());
    hostNode->enqueueOutgoingPacket(newPacket);
    return;
  }

  //REACHABILITY
  else if (  _packet->type == PacketType::DATA || _packet->type == PacketType::CONTROL_1 || _packet->type == PacketType::CONTROL_2){
    if (_packet->dstId == hostNode->getId() || _packet->dstId == -1) {
      hostNode->dispatchPacketToApplication(_packet);
      reachability.insert(hostNode->getId());
      LogSystem::EventsLogOutput.log( LogSystem::dstReach, Scheduler::now(),hostNode->getId(),(int)_packet->type,_packet->flowId,_packet->flowSequenceNumber);
    }
  }
}

//amIOnRing
bool ProbaFloodingDynamicRingRoutingAgent::amIOnRing( PacketPtr _packet ) {
  auto trIt = amIonRingMap.find(_packet->transmitterId);
  if (trIt == amIonRingMap.end()) 
    return false;

  else {
    if ((trIt->second.controlSeqNoBig > trIt->second.controlSeqNoSmall) || (trIt->second.controlSeqNoBig != -1 && trIt->second.controlSeqNoSmall == -1)){
      //cout << hostNode->getId() << " onringof " << _packet->transmitterId << endl;
      return true;}
    else
      return false;
  }
}

//forwardDataPackets
void ProbaFloodingDynamicRingRoutingAgent::forwardDataPackets( PacketPtr _packet ) {

  if (hostNode->getId() == 0) return;

    float optimal_NbForwardersPerHop = 15.0;
    float ratio_RingNeighborsOverForwarders = 4.0;
    float optimal_NbRingNeighborsPerHop = optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders;
    if ( hostNode->getEstimatedNeighbours() > optimal_NbRingNeighborsPerHop )

	rangeSmall = sqrt ((- optimal_NbForwardersPerHop * ratio_RingNeighborsOverForwarders * pow(ScenarioParameters::getRangeBig(),2) / hostNode->getEstimatedNeighbours()) + pow(ScenarioParameters::getRangeBig(),2));

    else rangeSmall = 0;
     
    cout << hostNode->getId() << " deden " << hostNode->getEstimatedNeighbours() << " rangeSmall " << rangeSmall << endl;

  if (needToSendControl) { //before the very first data transmission
    PacketPtr controlBig(new Packet(PacketType::CONTROL_1, 101,hostNode->getId(),-1,-1,-1,-1));
    controlBig->setCommunicationRange(rangeBig);
    controlBig->setCtrl1SequenceNumber(hostNode->getNextCtrl1SequenceNumber());
    hostNode->enqueueOutgoingPacket(controlBig);
    PacketPtr controlSmall(new Packet(PacketType::CONTROL_2, 102,hostNode->getId(),-1,-1,-1,-1));
    controlSmall->setCommunicationRange(rangeSmall);
    controlSmall->setCtrl2SequenceNumber(hostNode->getNextCtrl2SequenceNumber());
    hostNode->enqueueOutgoingPacket(controlSmall);	
    needToSendControl = false;													     
  }
  PacketPtr packetClone(_packet->clone());  
  // Don't use the delayed forwarding
  hostNode->enqueueOutgoingPacket(packetClone);
  forwardedDataPackets++;

  datapacket_t datapacket;
  datapacket.first =  packetClone->srcId;
  datapacket.second = packetClone->srcSequenceNumber;    
  auto rangeIt = countAckMap.find(datapacket); //or datapacket, node computes rangeSmall for each new source, or for each new packet?
  if (rangeIt == countAckMap.end()) { 												       countAckMap.insert(pair<datapacket_t,distance_t>(datapacket,rangeSmall));
    auto timerEvent = new ProbaFloodingDynamicRingAckEvent(Scheduler::now(), this, packetClone); 
    Scheduler::getScheduler().schedule(timerEvent);
  }       
}

//processAckTimerEvent
void ProbaFloodingDynamicRingRoutingAgent::processAckTimerEvent(PacketPtr _packet) { //what is _packet??


  datapacket_t datapacket;
  datapacket.first =  _packet->srcId;
  datapacket.second = _packet->srcSequenceNumber;    
  auto rangeIt = countAckMap.find(datapacket); 
  //todo if not end
  cout << "time " << Scheduler::now() << " node " << hostNode->getId() << " src " << rangeIt->first.first << " seqno " << rangeIt->first.second  << " rangeSmall " << rangeIt->second << "\n" << endl; 
}
